function hist = cycleBethe(theta, W, cyc, Nconstr, varargin)

    %W=-W;
    
    p = inputParser;
    p.addRequired('theta', @isnumeric);
    p.addRequired('W');
    p.addRequired('cyc');
    p.addRequired('Nconstr');
        
    p.addParamValue('outerEConst', 0)
    p.addParamValue('gapThresh', 1e-4);
    p.addParamValue('lamThresh', 1e-6);
    p.addParamValue('initStep', 1);
    p.addParamValue('weakDualityFudge', 1e-10);
    p.addParamValue('maxIter', 200);
    
    % Solve the initial problem without constraints.
    p.addParamValue('lamInit', []);
    p.addParamValue('showIntermediate', false);
    p.addParamValue('showSubgrads', false);        
    
    % Choice of inner solvers, so long as they have interface
    % [mu, xi, twoMargs, energy, complexity] = f(theta, W, varargin)

    if all(vec(W >= 0))
        % Attractive problem; we can use poly-time approximation
        p.addParamValue('innerSolver', @solveBethe);
    else
        % At least one negative weight; must we integer programming
        p.addParamValue('innerSolver', @solveBetheMAP);
    end
    
    p.addParamValue('innerSolverOpts', {});
    
    p.parse(theta, W, cyc, Nconstr, varargin{:});
            
    o = p.Results;
    
    N = length(theta);
    W = triu(W, 1);
    
    

    if isempty(o.lamInit)
        lam = zeros(1, Nconstr);
    else        
        lam = o.lamInit;
    end

            
    % The dual should increase. Count the times it decreased instead. We
    % use the quantity to adjust stepsize [Rush12].    
    dualDecreaseIters = 0;        
    stepSz = o.initStep;    
    gap = 10000;
            
    dTheta = zeros(N, 1);
    dW     = zeros(N, N);
    
    % Helper functions for the loop
    function reparam(u, v, dt, dw)
        dTheta([u,v]) = dTheta([u,v]) + dt;        
        dW(u,v) = dW(u,v) + dw;
    end

    % Parameter absorption functions
    function reparamFEdge(F, u, v)
        lm = lam(F.idx);            
        reparam(u, v, -lm, 2 * lm);
    end

    function reparamCompEdge(F, u, v)
        lm = lam(F.idx);
        reparam(u, v, lm, -2 * lm);
        Econst = Econst + lm;      
    end

    % Subgradient computation functions
    function sumFEdge(F, u, v)
        mainSums(F.idx) = mainSums(F.idx) + mainDiag(u, v);
    end

    function sumOffEdge(F, u, v)
        offSums(F.idx)  = offSums(F.idx) + offDiag(u, v);
    end

    % Initialize the loop
    gap = Inf;
    oldLam = Inf;
    iter = 1;
        
    % If either convergence criterion met, we quit.
    while iter < o.maxIter && abs(gap) > o.gapThresh && norm(lam - oldLam, 1) > o.lamThresh
        % Initialize the absorption parameters        
        iterTicID = tic;
        
        Econst = 0;
        dTheta = zeros(N, 1);
        dW     = zeros(N, N);        
                
        forEachConstraintAndEdge(cyc, @reparamFEdge, @reparamCompEdge);
        
        % Solve \argmin_{\mu} L(\mu, \lambda).
        
        assert(all(~isnan(theta) & ~isinf(theta)), 'Theta went Nan/Inf!');        
        
        % Solve it.
        if o.showIntermediate
            theta + dTheta
            W + dW
        end
                
        [mu, xi, twoMargs, energy, complexity, lagrOpt] = o.innerSolver(theta + dTheta, W + dW, o.innerSolverOpts{:});
        if ~lagrOpt
            warning('Inner solver did not certify optimality. Dual values may be wrong.');
        end                
        
        % The returned minimum energy should be attained by the returned
        % minimizer mu and twoMargs.
        %Fbethe = bethe(theta + dTheta, W + dW, mu, twoMargs);        
        %assertElementsAlmostEqual(Fbethe, energy, 'absolute', o.epsilon)        
        
        % We want a symmetric xi
        xi = triu(xi, 1);
        xi = xi + xi';                
        
        if o.showIntermediate
            xi
            twoMargs
        end
        
        % Solve \argmax_{\lambda} L(\mu, \lambda) by subgradient ascent.
        
        % Adrian's decomposition of the term in the parentheses.
        %
        % The subgradient is just the stuff inside the constraint
        % parentheses--the violations.
        muMat    = bsxfun(@plus, mu, mu');
        mainDiag = 1 + 2*xi - muMat;
        offDiag  = muMat - 2*xi;                        
        
        mainSums = zeros(1, Nconstr);
        offSums  = zeros(1, Nconstr);
        
        % Each coordinate of the subgradient is just the amount of
        % constraint violation of that constraint. We use the same word and
        % variable to denote both concepts.
        forEachConstraintAndEdge(cyc, @sumFEdge, @sumOffEdge);
        subGrad = 1 - mainSums - offSums;

        
        if o.showIntermediate || o.showSubgrads
            lam
            subGrad
        end                
        
        viols = subGrad > 0;
        % Round the current iterate mu to the cycle polytope by finding the
        % convex combination \alpha \nu + (1 - \alpha) \mu with minimal
        % \alpha.
        %
        % \alpha is determined by a formula (WHAT NUMBER?)        
        
        alpha = -Inf * ones(length(subGrad), 1);
        
        for nc = 1:length(cyc)
            for nf = 1:length(cyc(nc).F)
                F = cyc(nc).F(nf);
                
                if viols(F.idx)                
                    assert(subGrad(F.idx) > 0);
                    
                    NCEdges = size(cyc(nc).edges, 1);
                    alpha(F.idx) = subGrad(F.idx) / ( NCEdges / 2 - 1 + subGrad(F.idx));                    
                    assert(alpha(F.idx) < 1, 'Alpha > 1; fail!');
                end                
            end
        end
        
        % In case everything was feasible, reverts to a no-op.
        % fracUnif means fraction of uniform.
        fracUnif = max(max(alpha), 0);
        muRounded       = fracUnif/2 + (1 - fracUnif) * mu;
        twoMargsRounded = fracUnif/4 + (1 - fracUnif) * twoMargs;
        
        %[twoMargsRounded, xiRounded] = wtTwoMargs(W + dW, muRounded);        
        primal = bethe(theta, W, muRounded, twoMargsRounded) + o.outerEConst;
        
        compSlackViol = sum(lam .* subGrad);
        primalFeasViol    = sum(subGrad(viols));
        
        Efinal = energy + Econst + o.outerEConst;                                
        %Efinal = energy - Econst;
        FBethe = bethe(theta, W, mu, twoMargs) + o.outerEConst;
        dual   = FBethe + compSlackViol;
        
        gap = primal - dual;

                
        if gap < -o.weakDualityFudge
            warning('Dual exceeded primal; this should only happen if the inner solve is not exact.');
        end        
        
        iterTime = toc(iterTicID);
        
        % TODO: Don't repeat this typing?
        fprintf(1, 'Iter %d, iterTime=%g, dual=%g, primal=%g, gap=%g, compSlackViol=%g, primalFeasViol=%g, fracUnif=%g, lagrOpt=%d\n', iter, iterTime, dual, primal, gap, compSlackViol, primalFeasViol, fracUnif, lagrOpt);
        hist(iter) = var2struct(Efinal, iterTime, FBethe, dual, primal, fracUnif, compSlackViol, primalFeasViol, mu, twoMargs, muRounded, twoMargsRounded, lam, complexity, lagrOpt, subGrad);
        
        % If a constraint is violated (subGrad > 0), then lambda should not
        % be zero. (Use the formula a => b <=> ~a | b.)
        %assert(all((subGrad <= 0) | (lam ~= 0)), 'Got violated constraint with zero cost.')
        %
        % Actually, this is somewhat of a fallacy. The lambda was old, while
        % the subgradients were just calculated.
        
        % Step size schedule in in [Rush12]; see [Komodakis11] for other ideas.
        if iter > 1 && dual < hist(iter-1).dual
            % We overstepped, so scale down our stepsize.
            fprintf(1, 'Dual decreased; decreasing step size\n');
            dualDecreaseIters = dualDecreaseIters + 1;
            stepSz = o.initStep / (1 + dualDecreaseIters);
        end
        
%        stepSz = o.initStep / sqrt(iter);        
               
        % lam is set for next iteration.        
        % Projected subgradient ascent (multipliers must be positive), see
        % [Komodakis11].
        oldLam = lam;        
        lam    = max(0, lam + stepSz * subGrad);
        iter   = iter + 1;
    end                       
end

