function [ bestFinal, hist ] = ddrun( graphType,theta,W,gurobiLogFile,Heuristics,uorb,EPSILONS, lpRelaxAtEnd )
%[ bestFinal,first,hist ] = ddrun( graphType,theta,W,gurobiLogFile,Heuristics,uorb,EPSILONS )
%   Runs dual decomp with MPLP in multiple phases
%   Initially with epsilon=EPSILONS(1) to completion
%   (the very first run is stored in first, which used lam=0, since may
%   want this as the Bethe opt on the local polytope; hist is entire history)
%   Then starts with that converged lambda, uses next epsilon to conv
%   Topology is Cn or Kn
%   Default is unbiased input theta, W convention
%   Use 'b' for biased input theta, W

ECONVERTTHRESH=0.001; % threshold for checking Energy conversion unbiased->biased

% KT: Nothing is used here...
first = [];

if nargin<4; gurobiLogFile='gurobiLog'; end
if nargin<5; Heuristics=0.05; end % for gurobi MIP solver, 0.05 is default, try 0.2?
if nargin<7; EPSILONS=[1,0.5,0.25,0.1]; end % Default EPSILONS
if nargin<8; lpRelaxAtEnd=false; end
nEPSILONS=length(EPSILONS);

% Check W is symmetric
if ~isequal(W,W'); error('W is not symmetric.\n'); beep; end

n=length(theta);
if n<3; error('n must be at least 3\n'); end
if size(W,1)~=n; error('Size mismatch between theta and W.\n'); beep; end
if ~isequal(diag(W), zeros(n,1)); error('W is not zero on main diagonal.\n'); end

% theta must be a column vec
if size(theta,1)==1; theta=theta'; end

% Unbiased->Biased th, W convention if needed
if nargin<6 || uorb=='u'
    [~, ~, ~, unprimedEnergy, ~, ~] = solveDAIUnbiased(theta, W);
    theta=theta-sum(W)'/2; % (w/2 0; 0 w/2) = (0 -w/2)' # (0 0; 0 w) # (0 -w/2) # w/2 
    Econst = -sum(vec(triu(W))) / 2; % Energy is MINUS theta.mu
    [~, ~, ~, primedEnergy, ~, ~]   = solveDAI(theta, W);
    fprintf('Orig unbiased E = %g; Econst = %g; New primedEnergy + Econst = %g\n', unprimedEnergy, Econst, primedEnergy + Econst);
    if abs(unprimedEnergy-primedEnergy-Econst)>ECONVERTTHRESH; warning('Conversion changed Energy\n'); beep; end
else
    Econst=0;
end
% Henceforth assume direct (biased) formulation.

% Generate set of cycle constraints
if graphType=='K'
    [cyc, Nconstr] = completeChordlessCycles(n); % complete graph Kn
elseif graphType=='C'
    [cyc, Nconstr] = cycleChordlessCycle(n); % Cn
elseif graphType=='A'
    [cyc, Nconstr] = chordlessCycles(W);
else
    error('First parameter must be ''C'' for cycle or ''K'' for complete graph or ''A''.\n');
end

% Simple heuristic initial step size
initStep = mean(abs(W(:)));

hist=cell(nEPSILONS,1); best=cell(nEPSILONS,1);
nextLam=zeros(1, Nconstr);
for run=1:nEPSILONS
    fprintf('Starting run %d with epsilon=%5.3f...\n',run,EPSILONS(run));
    hist{run}=cycleBethe(theta, W, cyc, Nconstr, 'outerEConst', Econst, 'initStep', initStep, 'lamInit', nextLam, 'innerSolverOpts', {'solver', 'gurobi', 'Heuristics', Heuristics, 'lprelax', true,  'epsilon', EPSILONS(run), 'LogFile', gurobiLogFile}, 'gapThresh', EPSILONS(run)/3);
    if run==nEPSILONS && ( hist{run}(end).lagrOpt==0 ) % last run did not return integer solution, redo with best lam using integer solver
        fprintf('Final run did not return integer solution, redo with best lam using integer solver.\n');
        best=summarizeDD(hist{run});
        nextLam=hist{run}(best.iBestDual).lam;
        hist{run}=cycleBethe(theta, W, cyc, Nconstr, 'outerEConst', Econst, 'initStep', initStep, 'lamInit', nextLam, 'innerSolverOpts', {'solver', 'gurobi', 'Heuristics', Heuristics, 'lprelax', lpRelaxAtEnd,  'epsilon', EPSILONS(run), 'LogFile', gurobiLogFile}, 'gapThresh', EPSILONS(run)/3);
    end
    best=summarizeDD(hist{run});
    nextLam=hist{run}(best.iBestDual).lam;
end
bestFinal=best;

% %histSmall = cycleBethe(theta, W, cyc, Nconstr, 'outerEConst', Econst, 'initStep', initStep, 'lamInit', nextLam, 'innerSolverOpts', {'solver', 'mplp',  'epsilon', 0.1});
% bestSmall = summarizeDD(histSmall);

end

