function hist = margBethe(theta, W, varargin)
% margBethe  Find the Bethe free energy on the local polytope
%
%   All parameters similar to cycleBethe, but we use lam for Lagrange
%   multipliers.
%
%   To make the summarizeDD methods work, the ODD terms of hist are for the
%   iterates and the EVEN terms of hist are avg.

    p = inputParser;
    p.addRequired('theta', @isnumeric);
    p.addRequired('W');
        
    p.addParamValue('outerEConst', 0)
    p.addParamValue('gapThresh', 1e-4);
    p.addParamValue('lamThresh', 1e-6);
    p.addParamValue('initStep', 1);
    p.addParamValue('weakDualityFudge', 1e-10);

    p.addParamValue('maxIter', 200);
    
    % Solve the initial problem without constraints.
    p.addParamValue('lamInit', []);
    p.addParamValue('showIntermediate', false);
    p.addParamValue('showSubgrads', false);        
    
    % Choice of inner solvers, so long as they have interface
    % [mu, xi, twoMargs, energy, complexity] = f(theta, W, varargin)
    p.addParamValue('innerSolver', @solveBetheMAP);
    p.addParamValue('innerSolverOpts', {'OutputFlag', 0, 'epsilon', 0.01});
    
    % Gurobi licensing only allows one process to call it.
    % NOTE: The semaphore must already have been created and posted before
    % the first call here.
    p.addParamValue('gurobiSemaphore', []);
    
    p.parse(theta, W, varargin{:});
            
    o = p.Results;
    
    N = length(theta);
    
    [iVec, jVec, wVec] = findUT(W);
    nEdges = length(iVec);
    
    if isempty(o.lamInit)
        lam = zeros(2, 2, nEdges);
    else        
        lam = o.lamInit;
    end
        
    % Initialize the loop
    gap = Inf;
    avg.gap = Inf;
    stepSz = o.initStep;        
    oldlam = Inf * ones(2, 2, nEdges);
    iter = 1;
    dualDecreaseIters = 0;
    lamNorm = Inf;
    
    avg.lam = 0;
    
    % Stopping condition is NOT zero subgradient.
    while iter < o.maxIter && abs(gap) > o.gapThresh && abs(avg.gap) > o.gapThresh
        % Initialize the absorption parameters        
        iterTicID = tic;
        
        Econst = 0;
        dTheta = zeros(N, 1);
        dW     = zeros(N, N);        

        % Reparameterize
        for ne = 1:nEdges
            i = iVec(ne);
            j = jVec(ne);            
            
            Econst = Econst - lam(1,1,ne);
            dTheta(i) = dTheta(i) + lam(2,1,ne) - lam(1,1,ne);
            dTheta(j) = dTheta(j) + lam(1,2,ne) - lam(1,1,ne);
            dW(i,j)   = dW(i,j) + lam(1,1,ne) + lam(2,2,ne) - lam(1,2,ne) - lam(2,1,ne);
        end
        
        avg.lam = (1 - 1/iter) .* avg.lam + lam ./ iter;
        
        avgEconst = 0;
        avgdTheta = zeros(N, 1);
        avgdW     = zeros(N, N);                        
        
        for ne = 1:nEdges
            i = iVec(ne);
            j = jVec(ne);            
            
            avgEconst = avgEconst - avg.lam(1,1,ne);
            avgdTheta(i) = avgdTheta(i) + avg.lam(2,1,ne) - avg.lam(1,1,ne);
            avgdTheta(j) = avgdTheta(j) + avg.lam(1,2,ne) - avg.lam(1,1,ne);
            avgdW(i,j)   = avgdW(i,j) + avg.lam(1,1,ne) + avg.lam(2,2,ne) - avg.lam(1,2,ne) - avg.lam(2,1,ne);
        end                
                
        % Solve it.
        if o.showIntermediate
            disp(['theta, W at current dual']);
            theta + dTheta
            W + dW
            
            disp(['theta, W at current *averaged* dual']);
            theta + avgdTheta
            W + avgdW
        end                
        
        % Solve \min_{\mu \in \cal{L}} \cal{F}_{\theta'} (\mu) }                
        [mu, xi, twoMargs, energy, complexity, lagrOpt] = o.innerSolver(theta + dTheta, W + dW, o.innerSolverOpts{:});
        
        if ~lagrOpt
            warning('Inner solver did not certify optimality. Dual values may be wrong.');
        end                
        
        % Solve \min_{\mu^K \in \cal{M}} ...
        [lamVal, muK] = minMuK(W, lam);
        
        % Solve the same Bethe problem at the averaged dual variables
        [avg.mu, ~, avg.twoMargs, avgenergy, avg.complexity, avg.lagrOpt] = o.innerSolver(theta + avgdTheta, W + avgdW, o.innerSolverOpts{:});
        [avglamVal, avgmuK] = minMuK(W, avg.lam);
        
        [avg.muRounded, avg.twoMargsRounded, ~, avg.roundDist] = roundToMarg(avg.mu, avg.twoMargs, W);
        avg.primal = bethe(theta, W, avg.muRounded, avg.twoMargsRounded) + o.outerEConst;                
        
        avg.FBethe = bethe(theta, W, avg.mu, avg.twoMargs) + o.outerEConst;
        avg.dual   = avg.FBethe + avglamVal;
        
        avg.gap    = avg.primal - avg.dual;                        
        
        % Henceforth, we continue calculations leading to the dual updates
        % with just the current iterate.
        subGrad = muK - twoMargs;
        
        avgsubGrad = avgmuK - avg.twoMargs;
                
        if o.showIntermediate || o.showSubgrads
            lam
            subGrad
        end
        
        % Rounded primal.
        % Calling Gurobi in more than one processes crashes, due to license
        % restrictions.
        if ~isempty(o.gurobiSemaphore)
            semaphore('wait', o.gurobiSemaphore);
        end
        
        [muRounded, twoMargsRounded, pRounded, roundDist] = roundToMarg(mu, twoMargs, W);
        
        if ~isempty(o.gurobiSemaphore)
            semaphore('post', o.gurobiSemaphore);            
        end        
        
        primal = bethe(theta, W, muRounded, twoMargsRounded) + o.outerEConst;        
        
        viols = subGrad ~= 0;
        % TODO: PRIMAL ROUNDING
                
        compSlackViol  = sum(vec(lam .* subGrad));
        primalFeasViol = sum(vec(subGrad(viols)));
        
        avgviols = avgsubGrad ~= 0;
        avg.compSlackViol  = sum(vec(avg.lam .* avgsubGrad));
        avg.primalFeasViol = sum(vec(avgsubGrad(avgviols)));
        
        
        Efinal = energy + Econst + o.outerEConst;                                
        avg.Efinal = avgenergy + avgEconst + o.outerEConst;
        
        FBethe = bethe(theta, W, mu, transEachMarg(twoMargs)) + o.outerEConst; % crazy thing, let's try it
        FBethePerturbed = bethe(theta + dTheta, W + dW, mu, twoMargs) + o.outerEConst;
        dual   = FBethe + compSlackViol;
        %dual   = FBethe + lamVal;
        %dual    = FBethePerturbed + lamVal;
        
        gap = primal - dual;
                
        % Duality gap could be as negative as twice the epsilon.
        % TODO: Change this line if 
        %fudge = -2 * o.innerSolverOpts.epsilon - o.weakDualityFudge;
        
        if gap < -o.weakDualityFudge
            warning('Duality gap was negative. This should not be more negative than -2*epsilon');
        end
                
        iterTime = toc(iterTicID);
        
        
        % If a constraint is violated (subGrad > 0), then lambda should not
        % be zero. (Use the formula a => b <=> ~a | b.)
        %assert(all((subGrad <= 0) | (lam ~= 0)), 'Got violated constraint with zero cost.')
        %
        % Actually, this is somewhat of a fallacy. The lambda was old, while
        % the subgradients were just calculated.
        
        % Step size schedule in in [Rush12]; see [Komodakis11] for other ideas.
        if iter > 1 && dual < hist(iter-1).dual
            % We overstepped, so scale down our stepsize.
            fprintf(1, 'Dual decreased; decreasing step size\n');
            dualDecreaseIters = dualDecreaseIters + 1;
            stepSz = o.initStep / (1 + dualDecreaseIters);
        end
        
%        stepSz = o.initStep / sqrt(iter);        
               
        % lam is set for next iteration.        
        % Projected subgradient ascent (multipliers must be positive), see
        % [Komodakis11].
        oldlam = lam;        
        lam    = lam + stepSz .* subGrad;

        % TODO: Don't repeat this typing?
        lamNorm = norm(vec(lam - oldlam), 1); 
        
        fprintf(1, ['Iter %d, iterTime=%g, dual=%g, primal=%g, gap=%g, roundDist=%g\n'...
            'avg.dual=%g, avg.primal=%g, avg.gap=%g, avg.roundDist=%g\n' ...
            'lamNorm=%g, compSlackViol=%g, primalFeasViol=%g, roundDist=%g, lagrOpt=%d\n'], ...
            iter, iterTime, dual, primal, gap, roundDist,...
            avg.dual, avg.primal, avg.gap, avg.roundDist, ...
            lamNorm, compSlackViol, primalFeasViol, roundDist, lagrOpt);
        
        % Must match the order of avg.
        hist(2*iter - 1) = var2struct(gap, lam, mu, twoMargs, complexity, lagrOpt, ...
            muRounded, twoMargsRounded, roundDist, primal, FBethe, dual, ...
            compSlackViol, primalFeasViol, Efinal, iterTime, lamNorm, subGrad);

        % Silliness to fit into the same strucutre
        avg.iterTime = nan;
        avg.lamNorm =  nan;       
        avg.subGrad = avgsubGrad;
        
        
        hist(2*iter) = avg;
        
        iter   = iter + 1;
    end

end

