function [muRounded, twoMargsRounded, pRounded, dist, alphaOpt, extremes] = roundToMarg(mu, twoMargs, W)
% roundToMarg  Round a solution to the marginal polytope.
%
%   [muRounded, twoMargsRounded, pRounded, dist] = roundToMarg(mu, twoMargs, W)
%
%   muRounded and twoMargsRounded are the answer on the marginal polytope
%   which minimizes L1 distance to (mu, twoMargs). pRounded is the
%   overcomplete vector in the marginal polytope from which muRounded and
%   twoMargsRounded were extracted.
%
%   Implements checkInMargPolytope LP, using Gurobi, extended for arbitrary
%   (not necessarily complete) graphs.

    % Matrix form of the LP:
    %
    % x = [ s^1, \ldots, s^d, \alpha_1, \ldots, \alpha_{2^n} ]'
    %
    % Objective:
    %   Coefficient for each s^j        = +1
    %
    % A matrix: (2*d + 1) x (d + 2^n)
    %
    % Rows j \in 1, \ldots, d:
    %   Coefficient  for s^j            = +1
    %   Coefficients for each \alpha_i  = v_i^j (Row j, col i)
    %   Sense                           = \geq
    %   RHS                             = p^j
    %
    % Rows j \in d + 1, \ldots, 2*d
    %   Coefficient  for s^j            = -1
    %   Coefficients for each \alpha_i  = v_i^j
    %   Sense                           = \leq
    %   RHS                             = p^j
    %
    % Row 2*d + 1
    %   Coefficients for each \alpha_1 = 1
    %   Sense                          = \eq
    %   RHS                            = 1
    
    [extremes, D, nEdges] = sparseMargExtreme(W);
    
    iVec = [];
    jVec = [];
    wVec = [];
    
    N     = length(mu);
    nRows = 2*D + 1;    
    
    oneMargs(1,:) = 1 - mu;
    oneMargs(2,:) = mu;

    % Squeeze into one vector
    p = [oneMargs(:) ; twoMargs(:)];        
        
    model.obj = zeros(1, D + 2^N);
    model.obj(1:D) = 1;
    model.rhs = [p ; p ; 1];
    model.sense = [repmat('>', 1, D) repmat('<', 1, D) '='];
    
    % The columns for which we put an alpha coordinate
    alphaIdxs = (D+1):(D+1 + 2^N - 1);
    nAlpha    = length(alphaIdxs);
    % Could use this to replace length(idxs)
    
    % Add the first group of rows
    ind = 1;
    for j = 1:D
        % Column indices to put a coefficient
        idxs   = [j alphaIdxs];
        coeffs = [1 extremes(j,:)];

        assert(length(idxs) == length(coeffs), 'Youfail.org');

        range = ind:(ind + nAlpha);
        iVec(range) = j;
        jVec(range) = idxs;
        wVec(range) = coeffs;

        ind = ind + length(idxs);
    end
    
    for j = 1:D
        % Column indices to put a coefficient
        idxs   = [j alphaIdxs];
        coeffs = [-1 extremes(j,:)];

        assert(length(idxs) == length(coeffs), 'Youfail.org');

        range = ind:(ind + nAlpha);
        iVec(range) = D + j;
        jVec(range) = idxs;
        wVec(range) = coeffs;

        ind = ind + length(idxs);
    end
    
    % The last constraint
    range = ind:(ind + nAlpha - 1);
    iVec(range) = nRows;
    jVec(range) = alphaIdxs;
    wVec(range) = 1;    
    
    model.A = sparse(iVec, jVec, wVec);
    params = struct('OutputFlag', 0);
    
    save('gurobiDebug');
    
    result = gurobi(model, params);
    assert(strcmp(result.status, 'OPTIMAL'), 'Rounding LP not solved to optimality!');
    
    dist = result.objval;
    alphaOpt = result.x(alphaIdxs);
    pRounded = sum(bsxfun(@times, alphaOpt', extremes), 2);
    
    [muRounded, twoMargsRounded] = extractMargs(N, nEdges, pRounded);   

end

