classdef BetheFrankWolfe < handle
    % BetheFrankWolfe
    %
    %   ONLY BINARY PAIRWISE!
    
    % TODO: Different options structure!
    
    properties
        % Problem data
        theta
        W
        N
        nEdges
        degs
        thetaOvercomplete
        obj
%         otherObj
        grad
        
        % Function inputs; set to a default
        linesearch = 'fminbnd' % DEFAULT
        linesearchUb = 1.0 % CHANGE if not fminbnd.
        fMinBndOpts = optimset('TolX', 1e-9);
        
        % Iteration state
        iter
        hist
        mu
        
        % Options
        % TODO: Move this crap into an options structure passed in through
        % the command line, populated with defaults. Look up the example of
        % optimset.
        convFun = true
        convX   = false
        TolFun  = 1e-6
        TolX    = 1e-6
        
        boundaryFudge = 0
        fudgeCount
        
        mkA
        mkB
        
        % Debug Options
        plotLine
        plotInit
        ph
        print = true;        
        iterLimit = 10000;
        
        % low water marks
        fMin = Inf
        muMin   
    end
    
    methods (Abstract)
        target = solveSubLP(g, this)
        initMAP(this, mix)        
    end
    
    methods
        function this = BetheFrankWolfe(theta, W)
            % BetheFrankWolfe  Construct an optimization model.
            this.theta = theta;
            this.W     = W;
            
            this.fudgeCount = 0;
            
            [this.mkA, this.mkB] = bpbound(theta, W, 1e-5);
            
            this.N = length(theta);
            this.degs = sum(W ~= 0, 1);
            this.nEdges = sum(this.degs) / 2;
            
            this.thetaOvercomplete = biasedToOvercomplete(theta, W);
            
            this.obj  = @(x) betheOvercomplete(x, this.thetaOvercomplete, this.degs, this.N, this.nEdges);
%             this.otherObj  = @(x) fortBetheOvercomplete(x, this.thetaOvercomplete, this.degs, this.N, this.nEdges);
            this.grad = @(x) gradBetheOvercomplete(x, this.thetaOvercomplete, this.degs, this.N, this.nEdges);                                    
                        
            this.print = true;            
        end
        
        function initUnif(this)
            this.mu = makeUnifOvercomplete(this.N, this.nEdges);
        end
        
        function tf = isConverged(this)
            % isConverged
            if this.iter <= 2
                tf = false;
            else
                dfIter = this.hist{this.iter-1}.f - this.hist{this.iter-2}.f;
                dmIter = this.hist{this.iter-1}.f - this.hist{this.iter-2}.f;                    
                
                tf = (~this.convX   || norm(dmIter, 1) < this.TolX) && ...
                     (~this.convFun || norm(dfIter, 1) < this.TolFun);                     
            end
            
        end
                
        function nextIter(this)      
            f = this.obj(this.mu);
%             otherF = this.otherObj(this.mu);   
%             assertElementsAlmostEqual(f, otherF);
            g = this.grad(this.mu);            
            
            % Evaluate the gradient dot each extreme point.
            target = this.solveSubLP(g);
            
            % Linesearch: 0 is current x, 1 is target.            
            arg  = @(t) (1 - t)*this.mu + t*target;
            line = @(t) this.obj(arg(t));
                        
            % TODO: Refactor these. (Only makes sense if we figure out a
            % different linesearch that doesn't fit this pattern.)
            if strcmp(this.linesearch, 'fminbnd')
                % HACK could not get the attribute linesearchUb to work.
                [t, tFMin, exitflag] = fminbnd(line, 0, 1.0, this.fMinBndOpts);
                assert(exitflag == 1, 'fminbnd did not converge.');
                
                % Exception: Don't move.
                if line(0) < tFMin
                    t = 0;
                    tFMin = line(0);
                end                    
            elseif strcmp(this.linesearch, 'fminbndOLD')
                [t, tFMin, exitflag] = fminbnd(line, 0, 1.0);
                assert(exitflag == 1, 'fminbnd did not converge.');                
            elseif strcmp(this.linesearch, 'exhaustive')
                error('Broken for now; fix lineStep');
%                 lineStep = o.exhaustiveStepSz / norm(mu - target, 2);
%                 %             mesh     = [0:lineStep:1 1];
%                 % TODO: Vectorize! For loop is JITed.
%                 %
%                 mesh = linspace(0, this.linesearchUb, o.exhaustiveSteps);
%                 nSteps   = length(mesh);
%                 fvals = zeros(nSteps, 1);
%                 for i = 1:nSteps
%                     fvals(i) = line(mesh(i));
%                 end
%                 
%                 [tFMin, iMin] = min(fvals);
%                 t = mesh(iMin);
%                 
%                 if t == 1
%                     warning('margBetheFrankWolfe:extreme', 'Hitting an extreme point. Gradient may diverge to infinity.');
%                 end                
            elseif strcmp(this.linesearch, 'direct')
                Problem.f = line;
                bounds    = [0 this.linesearchUb];
                opts.tol  = o.TolX * 1e-2; % tighter tolerance here
                opts.maxevals = 2000;
                opts.maxits   = 100;
                [tFMin, t, directHistory] = Direct(Problem,bounds,opts);
            else
                error('BetheFrankWolfe:linesearch', 'Unsupported linesearch method.');
            end
            
            if this.plotLine
                %             clf(ph);
                figure(this.ph);
                ax = gca;
                hold on;
                ezplot(ax, line, [0, 1]);
                plot(ax, t, tFMin, '-.or');
                title(sprintf('Iter = %d, min = (%g, %g)', this.iter, t, tFMin));
                hold off;
            end                        
            
            % g(x) = \left< x - s, \nabla f(x) \right>; Jaggi13 Eq. (2)
            % Note that duality gap may be meaningless for our nonconvex
            % objective.
            %
            % Substitute x -> mu, s -> target.
            gap = sum((this.mu - target) .* g);
            
            % UPDATE THE ITERATE.
            this.mu = arg(t);
            if any(this.mu == 0) || any(this.mu == 1)
                % Just fudge it
                this.fudgeCount = this.fudgeCount + 1;
                this.mu(this.mu == 0) = this.boundaryFudge;
                this.mu(this.mu == 1) = 1 - this.boundaryFudge;                
            end

            mu = this.mu; % for naming consistency
%             this.hist = [this.hist var2struct(f, g, mu, t, target, tFMin, gap)];
            this.hist{this.iter} = var2struct(f, g, mu, t, target, tFMin, gap);
            
            if this.print
                fprintf(1, 'Iter=%d, f=%.10g, |g|_1=%g, gap=%g, t=%g\n', this.iter, f, norm(g, 1), gap, t);
            end
                        
            this.iter = this.iter + 1;
        end
        
        function tf = onBoundary(this)
            tf = any(this.mu <= 0) || any(this.mu >= 1);
        end
        
        function [muMin, fMin, hist, converged] = run(this)
            this.iter = 1;
            % assume already initialized
            while ~this.isConverged() && this.iter < this.iterLimit && ~this.onBoundary()
                this.nextIter();                                                
            end

            converged = (this.iter < this.iterLimit) && ~this.onBoundary();
                                  
            % Convert to cell array
            hist = [this.hist{:}];
                        
            % Check for monotonicity
            [fMin, iMin] = min([hist.f]);    

            if hist(iMin).f ~= hist(end).f
                % 30 -> 20140213T022040 format
                saveFile = sprintf('cp/BetheFrankWolfe_nonmonotonic_%s.mat', datestr(now, 30));
                if this.print && (~strcmp(this.linesearch,'fminbnd')) % if using fminbnd then don't expect monotonic
                    warning(['BetheFrankWolfe was not monotonic. Example saved to ' saveFile]);
                    save(saveFile); % moved in here so only save if verbose, sometimes caused crashes
                end
                %save(saveFile);
            end
    
            muMin = hist(iMin).mu;                        
        end
        
    end
    
end
