function [extremes, D, nEdges] = sparseMargExtreme(W)
% sparseMarginalPolytope  Generate extreme points of marginal polytope
%
%   extremes: D x 2^N where D = 2*N + 4*nEdges
%   This is similar Figure 2-1 on pp. 17 of Sontag's thesis. Each column is
%   an extreme point. The first 2*N dimensions are the overcomplete one
%   marginals, and the next 4*nEdges dimensions are the overcomplete two
%   marginals. Each one marginal looks like [0; 1] or [1; 0] and each two
%   marginal looks like [0 0 ; 0 1], etc. They are columnized (:).
%
%   You can recover them by reshape(extremes(n,1:(2*N)), 2, N)
%                           reshape(extremes(n:(2*N+1):end), 2, 2, nEdges);


    N = size(W, 1);
    [iVec, jVec, wVec] = findUT(W);
    nEdges = length(iVec);
    
    D = 2*N + 4*nEdges;
    extremes = zeros(D, 2^N);
    
    % For each configuration
    for n = 1:2^N
        mu = dec2bin(n - 1, N) - '0' + 1; % To get 1 or 2.        
        
        twoMargs = zeros(2, 2, nEdges);        
        % Add the components of \delta_ij(x_i, x_j) that muC says are 1.
        for ne = 1:nEdges
            i = iVec(ne);
            j = jVec(ne);
            
            twoMargs(mu(i),mu(j),ne) = 1;
        end
        
        oneMargs = zeros(2, N);
        % If mu(i) = 1, then oneMargs(1,i) = 1.
        % If mu(i) = 2, then oneMargs(2,i) = 1.
        oneMargs(1, mu == 1) = 1;
        oneMargs(2, mu == 2) = 1;
        
        % Squeeze into one vector
        extremes(:,n) = [oneMargs(:) ; twoMargs(:)];
    end
end

