function [muBest, fBest, muHist, fHist, fHistStd] = minBethe(theta, W, varargin)
% minBethe  Compute the local minimum of the Bethe free energy
%
%   [mu, f] = minBethe(theta, W) finds local optima from 100 starting
%   points.
%
%   [mu, f] = minBethe(theta, W, 'mu0', mu0) find local optima by
%   perturbing mu0 Gaussianly, with parameter 'perturbSd' default 0.01.
%   All perturbations/random starts guaranteed to be inside MK bounds.
%
%   muBest, fBest are the minimizer and value of the HIGHEST f.
%   muHist, fHist stores all values (e.g. to plot as histogram if we're
%   interested).
%
%   fHistStd gives the standard deviation of the fs; if it's very close to
%   zero, then we only found one minimum.
%
%   If we're reasonably sure of the derivatives, we can remove the
%   DerivativeCheck for speed.
    
    p = inputParser;

    p.addRequired('theta', @isnumeric);
    p.addRequired('W');
    p.addParamValue('nTrials', 100);
    p.addParamValue('mu0', []);
    p.addParamValue('perturbSd', 0.01);
    p.addParamValue('parallelRestarts', false);
    
    p.parse(theta, W, varargin{:});
    o = p.Results;

    [A, B] = bpbound(theta, W);
    N = length(theta);

    opts = optimset('GradObj', 'on', 'Hessian', 'user-supplied', ...
        'HessFcn', @(t, u) betheHess(t, u, theta, W), 'Algorithm', 'interior-point', ...        
        'Display', 'off', ...
        'FinDiffType', 'central', 'DerivativeCheck', 'off');
    
    prPrototype.objective = @(t) betheWithDerivs(t, theta, W);
    prPrototype.solver    = 'fmincon';    
    % Robustification in case they're close.
    prPrototype.lb        = min(A, 1 - B);
    prPrototype.ub        = max(A, 1 - B);
    prPrototype.options   = opts;
    
    muHist = cell(N, 1);
    fHist  = zeros(N, 1);        
    
%    global t;               
%    outputHist = {};
    
%    % Store the iterations to analyze
%    function stop = outputCb(x,optimValues,state)
%        if strcmp(state, 'iter')
%            i = optimValues.iteration + 1;
%            outputHist{t}(i).x = x;
%            outputHist{t}(i).f = optimValues.fval;
%        end
%        stop = false;
%    end

%    opts = optimset('OutputFcn', @outputCb, 'algorithm', 'interior-point', 'Display', 'none');
    
    prs = cell(o.nTrials, 1);
    if o.parallelRestarts
        parfor t = 1:o.nTrials
            prs{t} = prPrototype;
            if isempty(o.mu0)
                % Draw one randomly from [A, 1 - B]
                prs{t}.x0 = unifrnd(prs{t}.lb, prs{t}.ub);
            else
                % Perturb mu0 by perturbSd, and then clip.
                prs{t}.x0 = o.mu0 + o.perturbSd * randn(N, 1);
                prs{t}.x0(prs{t}.x0 < prs{t}.lb) = prs{t}.lb;
                prs{t}.x0(prs{t}.x0 > prs{t}.ub) = prs{t}.ub;
            end

            [muHist{t}, fHist(t), exitflag] = fmincon(prs{t});
        end
    else
        for t = 1:o.nTrials
            prs{t} = prPrototype;
            if isempty(o.mu0)
                % Draw one randomly from [A, 1 - B]
                prs{t}.x0 = unifrnd(prs{t}.lb, prs{t}.ub);
            else
                % Perturb mu0 by perturbSd, and then clip.
                prs{t}.x0 = o.mu0 + o.perturbSd * randn(N, 1);
                prs{t}.x0(prs{t}.x0 < prs{t}.lb) = prs{t}.lb;
                prs{t}.x0(prs{t}.x0 > prs{t}.ub) = prs{t}.ub;
            end

            [muHist{t}, fHist(t), exitflag] = fmincon(prs{t});
        end
    end
    
    [~, iBest] = max(fHist);
    muBest = muHist{iBest};
    fBest  = fHist(iBest);
    
    fHistStd = std(fHist);

end

