function [mu, xi, twoMargs, energy, complexity, exact] = solveBetheMAP(theta, W, varargin)
% [mu, xi, twoMargs, energy, complexity, exact] = solveBetheMAP(theta, W, varargin)
%
%   Find (global?) minimum of the Bethe free energy.
%
%   Returns the singleton potentials in mu (mu(i) == P(x_i = 1)) and
%   pairwise (2x2) potentials in twoMargs, with xi(i,j) = P(x_i = x_j = 1)
%
%   Right now, parameterizes an ILP using our bounds. But this will seem
%   silly. We should use gradient information. What about alpha-BB?
%
%   complexity is a structure with fields solverTime, sumN, prodN, thisN
%   count the sum, product, and vector of mesh points.
%
%   NOTE: Energy is -log Z. But the outer (dual decomposition) loop
%   introduces some constant shift, which it takes care of.
%

    p = inputParser;
    p.addRequired('theta', @isnumeric);
    p.addRequired('W');    
    p.addParamValue('epsilon', 0.1); % It's gonna be slow...
    p.addParamValue('solver', 'gurobi');
    p.addParamValue('plotEntropy', false);
    p.addParamValue('lprelax', false);
    p.addParamValue('deleteTempFiles', true);
    p.addParamValue('method', 'adaptiveminsum');
    % Gurobi params
    p.addParamValue('OutputFlag', 1);
    p.addParamValue('LogFile', 'gurobiLog');
    p.addParamValue('MIPGapAbs',0.01); % only for MIP, try epsilon http://www.gurobi.com/documentation/5.6/reference-manual/mipgap2
    p.addParamValue('NodeMethod', 1); % 0=primal simplex, 1=dual simplex, 2=barrier; all for MIP http://www.gurobi.com/documentation/5.6/reference-manual/nodemethod#parameter:NodeMethod
    p.addParamValue('Heuristics', 0.05); % amount of time spent in MIP heuristics; larger means more explore, less exploit http://www.gurobi.com/documentation/5.6/reference-manual/heuristics#parameter:Heuristics
    p.addParamValue('MIPFocus', 0); % balance between finding new feas solutions and proving optimal; try 3 if obj bound moving v slowy http://www.gurobi.com/documentation/5.6/reference-manual/mipfocus#parameter:MIPFocus 
    % for new impAB method
    p.addParamValue('impAB', false);
    
    p.parse(theta, W, varargin{:});    
    o = p.Results;
    
    W = triu(W, 1) + triu(W, 1)';
        
    [gams, complexity] = bestMesh(theta, W, o.epsilon, false, o.impAB);
    
    if complexity.sumN > 1000
        warning('solveBetheMAP:complexity', 'High mesh complexity; may take a long time.');
        complexity
        for i=1:length(complexity.thisN)
            fprintf('Number of discrete points in dimension %3d: %6d\n',i,complexity.thisN(i));
        end
    end
    
    [D, newW, Vi, Vm]  = makeDiscreteMRF(theta, W, gams);
    N = length(theta);    
    L = cellfun(@length, D);        
    
    if strcmp(o.solver, 'gurobi')
        exact = true;
        
        % Directly solve multi-label (I)LP with Gurobi.
                        
        % We add the trivial (variable and edge) constraints; see
        % [Komodakis11], expressed in the Ax = b. Note that x is a BINARY
        % vector; the label for a given node is expressed as a 1-hot
        % subvector of x.
        
        % In particular, each label for each variable has its index in x,
        % as does each pairwise label. Instead of trying to calculate
        % indices, we'll just precompute them.
        %
        % iTheta{u}(k) is the index of the the kth label on node n.
        % iW{u,v}(k,l) is the index of (k,l) label on edge (u,v).
        
        iTheta = cell(N, 1);
        obj    = [];
        
        start = 1;
        for u = 1:N            
            iTheta{u} = start:(start + L(u) - 1);
            start = iTheta{u}(end) + 1;
            
            % Set the objective coefficients (which are just the singleton
            % potentials)
            obj(iTheta{u}) = D{u};
        end
        
        % iW is UT.
        iW = cell(N, N);
        for u = 1:N
            for v = (u+1):N
                if Vi(u,v)                
                    nLabels = L(u) * L(v);                
                    seq     = start:(start + nLabels - 1);
                    iW{u,v} = reshape(seq, L(u), L(v));                                

                    start = seq(end) + 1;

                    % Set the objective coefficients (which are just the
                    % pairwise potentials)
                    %
                    % TODO: Extract this; put it in a better place... move all
                    % this LP stuff to a different function so its easier to
                    % debug (you can enter some Toy problems)                
                    obj(seq) = Vm{Vi(u,v)}(:);
                end
            end
        end
        
        % Objective (each entry is just the value of its potential)        
                
        rhs = [];                        
        Atemp = struct('Ai', [], 'Aj', [], 'Aw', [], 'iA', 1);        
        
        iConstr = 1;        
        % Single label constraints
        for u = 1:N                        
            for k = 1:L(u)
                % The constraint (sum of the subvector of x for node n
                % equals 1)                
                %A(iConstr, iTheta{u}) = 1;
                Atemp = addARow(Atemp, iConstr, iTheta{u}, 1);
                                
                rhs(iConstr) = 1;
                
                iConstr = iConstr + 1;
            end
        end
        
        for u = 1:N
            for v = (u + 1):N
                if iW{u,v}
                    % Row-wise marginalization constraints                
                    for ku = 1:L(u)
                        % Fix label ku and marginalize v
                        %A(iConstr, iW{u,v}(ku,:)) = 1;
                        Atemp = addARow(Atemp, iConstr, iW{u,v}(ku,:), 1);

                        % Set equality to ku label of u
                        %A(iConstr, iTheta{u}(ku)) = -1;
                        Atemp = addARow(Atemp, iConstr, iTheta{u}(ku), -1);
                        rhs(iConstr) = 0;                    

                        iConstr = iConstr + 1;                    
                    end

                    % Column-wise marginalization constraints
                    for kv = 1:L(v)
                        % Fix label kv and marginalize u
                        %A(iConstr, iW{u,v}(:,kv)) = 1;
                        Atemp = addARow(Atemp, iConstr, iW{u,v}(:,kv), 1);

                        % Set equality to the kv label of v
                        %A(iConstr, iTheta{v}(kv)) = -1;
                        Atemp = addARow(Atemp, iConstr, iTheta{v}(kv), -1);

                        rhs(iConstr) = 0;

                        iConstr = iConstr + 1;
                    end
                end
            end
        end
        
        nVars = length(obj);
        A = sparse(Atemp.Ai, Atemp.Aj, Atemp.Aw);
        %assertElementsAlmostEqual(A, Arecon);
        
        % Now do the optimization
        sense = repmat('=', 1, iConstr - 1);
        
        if o.lprelax
            ub = ones(nVars, 1);
            vtype = repmat('C', 1, nVars);
            model = var2struct(A, obj, rhs, vtype, sense, ub);                        
        else
            vtype = repmat('B', 1, nVars);            
            model = var2struct(A, obj, rhs, vtype, sense);
        end
                
        % HERE
        %params = struct('OutputFlag', 0);
        % http://www.gurobi.com/documentation/5.6/reference-manual/outputflag#parameter:OutputFlag
        % defaults to log to file and screen
%     p.addParamValue('LogFile', 'gurobiLog');
%     p.addParamValue('MIPGapAbs',0.01); % only for MIP, try epsilon http://www.gurobi.com/documentation/5.6/reference-manual/mipgap2
%     p.addParamValue('NodeMethod', 1); % 0=primal simplex, 1=dual simplex, 2=barrier; all for MIP http://www.gurobi.com/documentation/5.6/reference-manual/nodemethod#parameter:NodeMethod
%     p.addParamValue('Heuristics', 0.05); % amount of time spent in MIP heuristics; larger means more explore, less exploit http://www.gurobi.com/documentation/5.6/reference-manual/heuristics#parameter:Heuristics
%     p.addParamValue('MIPFocus', 0); % balance between finding new feas solutions and proving optimal; try 3 if obj bound moving v slowy http://www.gurobi.com/documentation/5.6/reference-manual/mipfocus#parameter:MIPFocus 
     
        %params = struct('LogFile', 'MIPGapAbs=eps_Heuristics=0.2_gurobiLog', 'MIPGapAbs', o.epsilon, 'Heuristics', 0.2); 
        % Using MIPGapAbs of epsilon, seemed to work well
        params = struct('LogFile', o.LogFile, 'MIPGapAbs', o.epsilon, 'NodeMethod', o.NodeMethod, 'Heuristics', o.Heuristics, 'MIPFocus', o.MIPFocus, 'OutputFlag', o.OutputFlag); 
        
        result = gurobi(model, params);
        assert(strcmp(result.status, 'OPTIMAL'), '(I)LP not solved to optimality!');
        energy = result.objval;
        
        complexity.solverTime = result.runtime;        
        
        % Recover the (integer) labels
        if o.lprelax
            % Recover marginals. If we were integral, we'd recover it!
            mu = zeros(N, 1);                
            S  = zeros(N, 1);
            for u = 1:N
                labelMix = result.x(iTheta{u});
%                assertElementsAlmostEqual(sum(labelMix), 1);
                mu(u) = sum(gams{u} * labelMix);
                
                % Nonzero entropy of the mixture means the LP relaxation
                % was not tight.
                S(u) = -sum(xlogx(labelMix));
                if S(u) ~= 0
                    fprintf(1, 'LP Relaxation: label mixture entropy for variable %d is %g\n', u, S(u));
                    if o.plotEntropy                    
                        figure;
                        plot(labelMix);
                        title(sprintf('Var %d; Entropy = %g', u, S(u)));
                    end
                end
            end
            
            [twoMargs, xi] = wtTwoMargs(W, mu);
            
            if sum(S) ~= 0
                exact = false;
%                warning('LP relaxation was not tight; total entropy of labels is %g\n', sum(S));
            end                
            
            % Early exit
            return;
            
        else        
            % Decode the integral answer
            label(N) = 0;
            for u = 1:N
                k = find(result.x(iTheta{u}));
                assert(length(k) == 1, 'Single label constraint failed!');
                label(u) = k;
            end                
        end
        
    else
        error('solveBetheMAP:solver', 'Solver %s not recognized', o.algo);
    end
    
    % Recover marginals        
    mu       = zeros(N, 1);

    for u = 1:N
        assert(length(gams{u}) == L(u) && length(D{u}) == L(u), 'marginal, energy, and label mismatch');
        mu(u) = gams{u}(label(u));
    end

    [twoMargs, xi] = wtTwoMargs(W, mu);
        

end

